########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.private_xml` – Private XML Storage support (:xep:`0049`)
#######################################################################

This module provides support for storing and retrieving private XML data on
the server as per :xep:`Private XML Storage <49>`.

.. autoclass:: PrivateXMLService

To register payload XSOs for private storage :class:`Query` is
exposed:

.. autoclass:: Query

"""

from .xso import Query  # NOQA
from .service import PrivateXMLService  # NOQA
